package ua.com.dxrkness.service;

import jakarta.servlet.http.Part;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import software.amazon.awssdk.core.sync.RequestBody;
import software.amazon.awssdk.services.s3.S3Client;
import software.amazon.awssdk.services.s3.model.Bucket;
import software.amazon.awssdk.services.s3.model.ObjectIdentifier;
import software.amazon.awssdk.services.s3.model.S3Object;
import ua.com.dxrkness.function.Action;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.net.URI;
import java.nio.channels.Channels;
import java.util.List;
import java.util.Optional;
import java.util.function.Consumer;
import java.util.function.Function;

@Service
public class S3Service {
    private final S3Client client;
    private final String bucketName;

    public S3Service(S3Client client,
                     @Value("${s3.bucket-name}") String bucketName) {
        this.client = client;
        this.bucketName = bucketName;
    }

    public void createDirectory(URI path) {
        final var pathString = Optional.of(path)
                .map(URI::toString)
                .filter(p -> p.charAt(p.length() - 1) == '/')
                .orElse(path.toString() + '/');

        Action createDir = () -> client.putObject(b -> b
                        .bucket(bucketName)
                        .key(pathString)
                        .build(),
                RequestBody.empty());
        Action waitForCreation = () -> client.waiter()
                .waitUntilObjectExists(b -> b
                        .bucket(bucketName)
                        .key(pathString)
                        .build());

        createDir.andThen(waitForCreation).execute();
    }

    public void delete(URI path) {
        Function<Void, List<ObjectIdentifier>> listAllObjectInDirectory = (ignored) ->
                client.listObjectsV2(b -> b
                                .prefix(path.toString())
                                .bucket(bucketName)
                                .build())
                        .contents().stream()
                        .map(S3Object::key)
                        .map(key -> ObjectIdentifier.builder()
                                .key(key)
                                .build())
                        .toList();
        Consumer<List<ObjectIdentifier>> deleteObjects =
                keys -> client.deleteObjects(b -> b
                        .bucket(bucketName)
                        .delete(b1 -> b1
                                .objects(keys)));
        Action waitForDeletion = () -> client.waiter()
                .waitUntilObjectNotExists(b -> b
                        .bucket(bucketName)
                        .key(path.toString())
                        .build());

        listAllObjectInDirectory
                .andThen(keys -> {
                    deleteObjects.accept(keys);
                    return null;
                })
                .andThen(n -> {
                    waitForDeletion.execute();
                    return null;
                })
                .apply(null);
    }

    public List<String> listing(URI path) {
        return client
                .listObjectsV2(b -> b
                        .bucket(bucketName)
                        .prefix(path.toString())
                        .build())
                .contents()
                .stream()
                .map(S3Object::key)
                .filter(filePath -> !filePath.equals(path.toString()))
                .toList();
    }

    public void uploadFile(Part file, URI destinationPath) throws IOException {
        client.putObject(b -> b
                        .bucket(bucketName)
                        .key(destinationPath
                                .resolve(file.getSubmittedFileName().replace("\s", "_"))
                                .toString()),
                RequestBody.fromInputStream(file.getInputStream(), file.getSize()));
    }

    public BufferedInputStream downloadFile(URI key) {
        return new BufferedInputStream(
                Channels.newInputStream(
                        Channels.newChannel(client.getObject(b -> b
                                .bucket(bucketName)
                                .key(key.toString())))));
    }

    public List<String> listBuckets() {
        return client.listBuckets()
                .buckets()
                .stream()
                .map(Bucket::name)
                .toList();
    }
}